// Copyright 2007 Michiel Helvensteijn
// 
// This file is part of Timekeeper.
// 
// Timekeeper is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// Timekeeper is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Timekeeper; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

#ifndef ACTIVITY_H
#define ACTIVITY_H

#include <QString>
#include <QDateTime>
#include <QTime>

/**
 * Represents an activity (from a start to a stop) in the timekeeper model.
 * 
 * @author Michiel Helvensteijn
 * @date 15-05-2007
 */
class Activity {
	public:
		
		/**
		 * Standard constructor, creating an uninitiated activity.
		 */
		Activity();
		
		/**
		 * Activity constructor that can take the category, starting time, end time and note. Can also be called without the end time and the note.
		 * 
		 * @param category The category of the activity.
		 * @param start The start time of the activity.
		 * @param end The end time of the activity.
		 * @param note Some extra information about the activity.
		 */
		Activity(const QString & category, const QDateTime & start, const QDateTime & end, const QString & note = QString());
		
		/**
		 * Get the category of this activity.
		 * 
		 * @return The category of this activity
		 */
		const QString & getCategory() const;
		
		/**
		 * Get the start time of this activity.
		 * 
		 * @return The start time of this activity
		 */
		const QDateTime & getStart() const;
		
		/**
		 * Get the end time of this activity.
		 * 
		 * @return The end time of this activity
		 */
		const QDateTime & getEnd() const;
		
		/**
		 * Get the duration of this activity.
		 * 
		 * @return The duration of this activity.
		 */
		QTime getDuration() const;
		
		/**
		 * Get the extra information about this activity.
		 * 
		 * @return The extra information about this activity
		 */
		const QString & getNote() const;
		
		/**
		 * Set the category of this activity.
		 * 
		 * @param category The new category for this activity
		 */
		void setCategory(const QString & category);
		
		/**
		 * Set the starting time of this activity.
		 * The duration (not the end-time) will change
		 * as well to create a consistent state.
		 * 
		 * @param start The new starting time for this activity
		 */
		void setStart(const QDateTime & start);
		
		/**
		 * Set the end time of this activity.
		 * The duration (not the start-time) will change
		 * as well to create a consistent state.
		 * 
		 * @param end The new end time for this activity
		 */
		void setEnd(const QDateTime & end);
		
		/**
		 * Set the duration of this activity.
		 * The end-time (not the start-time) will change
		 * as well to create a consistent state.
		 * 
		 * @param duration The new duration of this activity in seconds
		 */
		void setDuration(QTime duration);
		
		/**
		 * Set the extra information of this activity.
		 * 
		 * @param note The new extra information for this activity
		 */
		void setNote(const QString & note);
		
	private:
		
		QString _category;
		QDateTime _start;
		QDateTime _end;
		QString _note;
};

#endif // ACTIVITY_H
