// Copyright 2007 Michiel Helvensteijn
// 
// This file is part of Timekeeper.
// 
// Timekeeper is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// Timekeeper is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Timekeeper; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

#ifndef LOGFILEHANDLER_H
#define LOGFILEHANDLER_H

#include "exceptions.h"

#include <QString>

/**
 * Base class for any class that keeps and manipulates a log-file.
 * This means that the filename must be known before the first
 * save-action.
 * 
 * @author Michiel Helvensteijn
 * @date 23-05-2007
 */
class LogfileHandler {
	public:
		
		virtual ~LogfileHandler() { }
		
		/**
		 * Set the filename for subsequent save or load actions.
		 * If there is a problem (for example, with permissions)
		 * an exception will be thrown in the correct function.
		 * Not in this one.
		 * 
		 * @param filename The file to 
		 */
		void setFilename(QString filename);
		
		/**
		 * Get the filename used for subsequent save or load actions.
		 * 
		 * @return The filename.
		 */
		QString getFilename() const;
		
		/**
		 * Load the logfile specified with setFilename().
		 * 
		 * @throw NoSuchFile If the file does not exist.
		 * @throw NoPermission If you have no permission to read the file.
		 * @throw InvalidFile If the file is not a valid Timekeeper log.
		 */
		virtual void load() throw (NoSuchFile, NoPermission, InvalidFile) =0;
		
		/**
		 * Import another logfile specified with \a filename .
		 * 
		 * @param filename The logfile to import.
		 * @throw NoSuchFile If the file does not exist.
		 * @throw NoPermission If you have no permission to read the file.
		 * @throw InvalidFile If the file is not a valid Timekeeper log.
		 */
		virtual void import(QString filename) throw (NoSuchFile, NoPermission, InvalidFile) =0;
		
		/**
		 * Save the log to the file specified with setFilename().
		 * 
		 * @throw NoPermission If you have no permission to write to the file.
		 * @throw UnknownError If an unknown error occurred during writing.
		 */
		virtual void save() const throw (NoPermission, UnknownError) =0;
		
	private:
		
		QString _filename;
};

#endif // LOGFILEHANDLER_H
