// Copyright 2007 Michiel Helvensteijn
// 
// This file is part of Timekeeper.
// 
// Timekeeper is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// Timekeeper is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Timekeeper; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

#include "mainwindow.h"

#include <QInputDialog>
#include <QApplication>
#include <QFileDialog>
#include <QMessageBox>
#include <QCloseEvent>
#include <QStatusBar>
#include <QSettings>
#include <QInputDialog>
#include <QHeaderView>

#include "parsernode.h"
#include "geticon.h"

MainWindow::MainWindow(QWidget * parent, Qt::WFlags flags) : QMainWindow(parent, flags) {
	QCoreApplication::setOrganizationName("Michiel Helvensteijn");
	QCoreApplication::setOrganizationDomain("michiel.eu");
	QCoreApplication::setApplicationName("Timekeeper");
	
	setupUi(this);
	createView();
	makeConnections();
	
	QSettings settings;
	if (settings.contains("mainwindow/toolbars")) {
		restoreState(settings.value("mainwindow/toolbars").toByteArray());
	}
	
	if (qApp->arguments().count() <= 1 || !openFile(qApp->arguments().at(1))) {
		bool worked = false;
		do {
			int answer = QMessageBox::question(this, tr("Timekeeper"), tr("Do you want to start a new log or open an existing one?"), tr("Exit"), tr("Open an existing log"), tr("Start a new log"), 0, 0);
			if (answer == 0) {
				exit(0);
			} else if (answer == 1) {
				worked = openFile();
			} else {
				worked = newFile();
			}
		} while (worked == false);
	}
	
	initialize();
}

MainWindow::~MainWindow() {
	QSettings settings;
	settings.setValue("mainwindow/toolbars",     saveState());
	settings.setValue("systray/visible",         _systray->isVisible());
	settings.setValue("mainwindow/sortcolumn",   _view->header()->sortIndicatorSection());
	settings.setValue("mainwindow/sortorder",    _view->header()->sortIndicatorOrder());
	settings.setValue("mainwindow/column0width", _view->columnWidth(0));
	settings.setValue("mainwindow/column1width", _view->columnWidth(1));
	settings.setValue("mainwindow/column2width", _view->columnWidth(2));
	settings.setValue("mainwindow/column3width", _view->columnWidth(3));
	settings.setValue("mainwindow/column4width", _view->columnWidth(4));
}

void MainWindow::closeEvent(QCloseEvent* event) {
	if (_systray->isVisible()) {
		hide();
		event->ignore();
		return;
	}
	
	event->accept();
}

void MainWindow::createView() {
	QSettings settings;
	
	// Set up view-filter-model connection
	_model = new TimekeeperModel(this);
	_filter = new TimekeeperFilter(this);
	_filter->setSourceModel(_model);
	_view->setModel(_filter);
	
	// Restore treeview things
	_view->sortByColumn(settings.value("mainwindow/sortcolumn", "1").toInt(), (Qt::SortOrder)settings.value("mainwindow/sortorder", "0").toInt());
	if (settings.contains("mainwindow/column0width")) {
		_view->setColumnWidth(0, settings.value("mainwindow/column0width").toInt());
		_view->setColumnWidth(1, settings.value("mainwindow/column1width").toInt());
		_view->setColumnWidth(2, settings.value("mainwindow/column2width").toInt());
		_view->setColumnWidth(3, settings.value("mainwindow/column3width").toInt());
		_view->setColumnWidth(4, settings.value("mainwindow/column4width").toInt());
	}
	
	// Set statusbar
	_selection = new QLabel(this);
	_totalDuration = new QLabel(this);
	statusBar()->addWidget(_selection, 1);
	statusBar()->addWidget(_totalDuration, 0);
	
	// Set toolbar things
	_newCategory = new QComboBox(_activity);
	_newCategory->setEditable(true);
	_newCategory->setSizeAdjustPolicy(QComboBox::AdjustToContents);
	_newCategory->lineEdit()->setMaxLength(30);
	_newCategory->setMinimumContentsLength(10);
	_newCategory->setDuplicatesEnabled(false);
	_newCategory->setInsertPolicy(QComboBox::NoInsert);
	_activity->insertWidget(_actionStart, _newCategory);
	selectionChanged();
	
	// Set This-Period menu
	_thisPeriodMenu = new QMenu(this);
	_thisPeriodMenu->addAction(_actionThisDay);
	_thisPeriodMenu->addAction(_actionThisWeek);
	_thisPeriodMenu->addAction(_actionThisMonth);
	_thisPeriodMenu->addAction(_actionThisYear);
	_thisPeriod->setMenu(_thisPeriodMenu);
	
	// System tray icon menu
	_systrayMenu = new QMenu(this);
	_systrayMenu->setWindowTitle("Timekeeper"); // In case this will ever be implemented in QMenu
	_systrayMenu->addAction(_actionStartSpecify);
	_systrayMenu->addAction(_actionStop);
	_systrayMenu->addSeparator();
	_systrayMenu->addAction(_actionNew);
	_systrayMenu->addAction(_actionOpen);
	_systrayMenu->addSeparator();
	_systrayMenu->addAction(_actionMinimize);
	_systrayMenu->addAction(_actionQuit);
	
	// System tray icon
	_systray = new QSystemTrayIcon(this);
	_systray->setIcon(getIcon("images/idle_32x32.png"));
	_systray->setContextMenu(_systrayMenu);
	_systray->setVisible(settings.value("systray/visible", true).toBool());
	_actionShowSystray->setChecked(_systray->isVisible());
	
	// Show-menu
	_menubar->addMenu(createPopupMenu());
}

void MainWindow::makeConnections() {
	// File menu
	connect(_actionNew,    SIGNAL(triggered()), this, SLOT(newFile())    );
	connect(_actionOpen,   SIGNAL(triggered()), this, SLOT(openFile())   );
	connect(_actionImport, SIGNAL(triggered()), this, SLOT(importFile()) );
	connect(_actionReport, SIGNAL(triggered()), this, SLOT(report())     );
	connect(_actionClose,  SIGNAL(triggered()), this, SLOT(close())      );
	connect(_actionQuit,   SIGNAL(triggered()), qApp, SLOT(quit())       );
	
	// Show menu
	connect(_actionShowSystray, SIGNAL(toggled(bool)), _systray, SLOT(setVisible(bool)));
	
	// The actions
	connect(_actionStart,        SIGNAL(triggered()), this, SLOT(startActivity())        );
	connect(_actionStartSpecify, SIGNAL(triggered()), this, SLOT(startActivitySpecify()) );
	connect(_actionStop,         SIGNAL(triggered()), this, SLOT(stopActivity())         );
	connect(_actionDelete,       SIGNAL(triggered()), this, SLOT(deleteActivities())     );
	
	connect(_view->selectionModel(), SIGNAL(selectionChanged(const QItemSelection&, const QItemSelection&)), this, SLOT(selectionChanged()));
	connect(_filter,                 SIGNAL(modelReset()),                                _view->selectionModel(), SLOT(clearSelection()));
	
	// Filter: From view to object
	connect(_category,        SIGNAL(activated(const QString &)), _filter, SLOT(setCategory(const QString &)) );
	connect(_from,            SIGNAL(dateChanged(const QDate &)), _filter, SLOT(setFrom(const QDate &))       );
	connect(_to,              SIGNAL(dateChanged(const QDate &)), _filter, SLOT(setTo(const QDate &))         );
	connect(_previous,        SIGNAL(clicked()),                  _filter, SLOT(setToPreviousPeriod())        );
	connect(_next,            SIGNAL(clicked()),                  _filter, SLOT(setToNextPeriod())            );
	connect(_actionThisDay,   SIGNAL(triggered()),                _filter, SLOT(thisDay())                    );
	connect(_actionThisWeek,  SIGNAL(triggered()),                _filter, SLOT(thisWeek())                   );
	connect(_actionThisMonth, SIGNAL(triggered()),                _filter, SLOT(thisMonth())                  );
	connect(_actionThisYear,  SIGNAL(triggered()),                _filter, SLOT(thisYear())                   );
	
	// Filter: From object to view
	connect(_filter, SIGNAL(categoryChanged(const QString &)), _category, SLOT(setEditText(const QString &)) );
	connect(_filter, SIGNAL(fromChanged(const QDate &)),       _from,     SLOT(setDate(const QDate &))       );
	connect(_filter, SIGNAL(toChanged(const QDate &)),         _to,       SLOT(setDate(const QDate &))       );
	
	// From options to selection-view
	connect(_model, SIGNAL(categoriesChanged(const QStringList &)), this, SLOT(updateCategoryList(const QStringList &)) );
	
	// From model/filter to statusbar
	connect(_filter, SIGNAL(dataChanged(const QModelIndex&, const QModelIndex&)), this, SLOT(refreshStatusbar()) );
	connect(_filter, SIGNAL(modelReset()),                                        this, SLOT(refreshStatusbar()) );
	connect(_filter, SIGNAL(rowsInserted(const QModelIndex&, int, int)),          this, SLOT(refreshStatusbar()) );
	connect(_filter, SIGNAL(rowsRemoved(const QModelIndex&, int, int)),           this, SLOT(refreshStatusbar()) );
	
	// Systray
	connect(_systray,        SIGNAL(activated(QSystemTrayIcon::ActivationReason)), this, SLOT(sysTrayActivation(QSystemTrayIcon::ActivationReason)) );
	connect(_actionMinimize, SIGNAL(triggered()),                                  this, SLOT(toggleSysTrayMinimize())                              );
}

bool MainWindow::newFile() {
	QString newFilename = "";
	
	newFilename = QFileDialog::getSaveFileName(this, tr("Specify the new filename"), QSettings().value("files/savedir", "").toString(), tr("Timekeeper logs") + QString(" (*.tkl);;") + tr("All filetypes") + QString(" (*.*)"));
	if (newFilename == "") return false;
	QSettings().setValue("files/savedir", QFileInfo(newFilename).absolutePath());
	
	_model->setFilename(newFilename);
	_model->clear();
	
	updateCategoryList(_model->getCategories());
	
	return true;
}

bool MainWindow::openFile(QString filename) {
	try {
		if (filename == "") {
			filename = QFileDialog::getOpenFileName(this, tr("Open file"), QSettings().value("files/savedir", "").toString(), tr("Timekeeper logs") + QString(" (*.tkl);;") + tr("All filetypes") + QString(" (*.*)"));
			if (filename == "") return false;
			
			QSettings().setValue("files/savedir", QFileInfo(filename).absolutePath());
			
			_model->setFilename(filename);
		} else {
			_model->setFilename(filename);
		}
		_model->load();
		
		updateCategoryList(_model->getCategories());
		
		return true;
	} catch (NoSuchFile) {
		QMessageBox::critical(this, tr("Open file"), tr("There is no such file."), QMessageBox::Ok, QMessageBox::NoButton);
	} catch (NoPermission) {
		QMessageBox::critical(this, tr("Open file"), tr("You have no permission to read that file."), QMessageBox::Ok, QMessageBox::NoButton);
	} catch (InvalidFile) {
		QMessageBox::critical(this, tr("Open file"), tr("The file has an invalid format."), QMessageBox::Ok, QMessageBox::NoButton);
	}
	
	return false;
}

bool MainWindow::importFile() {
	try {
		QString filename = QFileDialog::getOpenFileName(this, tr("Import file"), QSettings().value("files/savedir", "").toString(), tr("Timekeeper logs") + QString(" (*.tkl);;") + tr("All filetypes") + QString(" (*.*)"));
		if (filename == "") return false;
		
		if (QMessageBox::critical(this, tr("Import file"), tr("The changes caused by importing this file will automatically be saved.\nAre you sure you want to do this?"), QMessageBox::Yes | QMessageBox::No) == QMessageBox::Yes) {
			_model->import(filename);
			updateCategoryList(_model->getCategories());
			tryToSave();
			
			return true;
		}
	} catch (NoSuchFile) {
		QMessageBox::critical(this, tr("Import file"), tr("There is no such file."), QMessageBox::Ok, QMessageBox::NoButton);
	} catch (NoPermission) {
		QMessageBox::critical(this, tr("Import file"), tr("You have no permission to read that file."), QMessageBox::Ok, QMessageBox::NoButton);
	} catch (InvalidFile) {
		QMessageBox::critical(this, tr("Import file"), tr("The file has an invalid format."), QMessageBox::Ok, QMessageBox::NoButton);
	}
	
	return false;
}

void MainWindow::report() {
	QString templateFile;
	QString destinationFile;
	QString output;
	
	
	try {
		// Ask for template-filename
		
		templateFile = QFileDialog::getOpenFileName(NULL, tr("Specify the template file"), QSettings().value("files/reporttemplatedir", "").toString(), tr("Timekeeper template") + QString(" (*.tkt);;") + tr("All filetypes") + QString(" (*.*)"));
		if (templateFile == "") return;
		QSettings().setValue("files/reporttemplatedir", QFileInfo(templateFile).absolutePath());
		
		// Import template-file
		
		ParserNode* root = new ParserNode(ParserNode::Aggregation);
		
		QFile inputFile(templateFile);
		
		if (!inputFile.exists())
			throw NoSuchFile();
		
		if (!inputFile.open(QIODevice::ReadOnly))
			throw NoPermission();
		
		QTextStream inputStream(&inputFile);
		
		root->parse(ParserNode::tokenize(inputStream.readAll()));
		
		inputFile.close();
		
		// Output to string
		
		TimekeeperFilter* filter = new TimekeeperFilter;
		filter->setSourceModel(_model);
		filter->setAll();
		filter->sort(1);
		
		QMap<QString, TimekeeperFilter*> sets;
		sets.insert("all", filter);
		
		QTextStream outputStream(&output);
		root->build(outputStream, sets);
		
		delete filter;
		
		// Ask for report-filename
		
		destinationFile = QFileDialog::getSaveFileName(NULL, tr("Specify the destination file"), QSettings().value("files/reportdestdir", "").toString());
		if (destinationFile == "") return;
		QSettings().setValue("files/reportdestdir", QFileInfo(destinationFile).absolutePath());
		
		// Output string to report-file
		
		QFile outputFile(destinationFile);
		
		if (!outputFile.open(QIODevice::WriteOnly))
			throw NoPermission();
		
		if (outputFile.write(output.toStdString().c_str(), output.length()) == -1) {
			outputFile.close();
			throw UnknownError();
		}
		
		outputFile.close();
	} catch (TemplateError e) {
		QMessageBox::critical(NULL, tr("Generate report"), tr("There was an error in the template file.\n") + e.message());
		return;
	} catch (NoSuchFile) {
		QMessageBox::critical(NULL, tr("Generate report"), tr("The specified file does not exist."));
		return;
	} catch (NoPermission) {
		QMessageBox::critical(NULL, tr("Generate report"), tr("You do not have permission to read the specified file."));
		return;
	} catch (UnknownError) {
		QMessageBox::critical(NULL, tr("Generate report"), tr("An unknown error occurred while saving the report file."));
		return;
	}
}

void MainWindow::startActivitySpecify() {
	bool ok;
	QString category = QInputDialog::getItem(this, tr("Start activity"), tr("Specify the category:"), _model->getCategories(), _newCategory->currentIndex(), true, &ok);
	
	if (ok) {
		startThisActivity(category);
	}
}

void MainWindow::startActivity() {
	startThisActivity(_newCategory->currentText());
}

void MainWindow::startThisActivity(const QString& category) {
	if (category.isEmpty()) {
		QMessageBox::critical(this, tr("Start activity"), tr("You can't use an empty category name.\nPlease try another name."));
	} else if (category == TimekeeperFilter::AllCategories) {
		QMessageBox::critical(this, tr("Start activity"), tr("You can't call a category '%1'.\nPlease try another name.").arg(TimekeeperFilter::AllCategories));
	} else {
		_model->startActivity(category);
		_model->addCategory(category);
		setIdleState(false);
		tryToSave();
	}
}

void MainWindow::stopActivity() {
	_model->stopCurrentActivity();
	setIdleState(true);
	tryToSave();
}

void MainWindow::deleteActivities() {
	int firstRow = _view->selectionModel()->selectedRows().first().row();
	int lastRow = _view->selectionModel()->selectedRows().last().row();
	_filter->removeRows(firstRow, lastRow - firstRow + 1);
	
	selectionChanged();
	tryToSave();
	setIdleState(!_model->activityRunning());
}

void MainWindow::selectionChanged() {
	_actionDelete->setEnabled(_view->selectionModel()->hasSelection());
	_actionDelete->blockSignals(!_view->selectionModel()->hasSelection());
}

void MainWindow::setIdleState(bool idle) {
	_actionStart->setEnabled(idle);
	_actionStart->blockSignals(!idle);
	_actionStartSpecify->setEnabled(idle);
	_actionStartSpecify->blockSignals(!idle);
	_actionStop->setEnabled(!idle);
	_actionStop->blockSignals(idle);
	_newCategory->setEnabled(idle);
	
	if (idle) {
		_systray->setIcon(getIcon("images/idle_32x32.png"));
		_systray->setToolTip(tr("No activity running"));
	} else {
		_systray->setIcon(getIcon("images/running_32x32.png"));
		_systray->setToolTip(tr("The '%1' activity is running").arg(_model->runningCategory()));
	}
}

void MainWindow::tryToSave() {
	try {
		_model->save();
	} catch (NoPermission) {
		QMessageBox::warning(this, tr("Error writing to log-file"), tr("You do not have permission to write to the log-file."));
	} catch (UnknownError) {
		QMessageBox::warning(this, tr("Error writing to log-file"), tr("An unknown error occurred while writing to the log-file."));
	}
}

void MainWindow::sysTrayActivation(QSystemTrayIcon::ActivationReason reason) {
	if (reason == QSystemTrayIcon::Trigger) {
		toggleSysTrayMinimize();
	}
}

void MainWindow::toggleSysTrayMinimize() {
	if (isVisible()) {
		hide();
		_actionMinimize->setText(tr("&Restore"));
		_actionMinimize->setToolTip(tr("Restore Timekeeper window"));
		_actionMinimize->setIconText(tr("Restore"));
	} else {
		show();
		//raise();
		//activateWindow();
		_actionMinimize->setText(tr("&Minimize"));
		_actionMinimize->setToolTip(tr("Minimize Timekeeper window to system tray"));
		_actionMinimize->setIconText(tr("Minimize"));
	}
}

void MainWindow::initialize() {
	QSettings settings;
	
	_filter->thisWeek();
	_filter->setCategory(TimekeeperFilter::AllCategories);
	setIdleState(!_model->activityRunning());
}

void MainWindow::refreshStatusbar() {
	_selection->setText(QString(" ") + _filter->getFrom().toString() + QString(" - ") + _filter->getTo().toString() + QString(" (") + _filter->getCategory() + QString(") "));
	
	unsigned int duration = _filter->totalDuration();
	unsigned int hours    = duration / (60 * 60);
	unsigned int minutes  = duration / 60 % 60;
	unsigned int seconds  = duration % 60;
	
	_totalDuration->setText(QString(" %1:%2:%3 ").arg(hours).arg(minutes, 2, 10, QChar('0')).arg(seconds, 2, 10, QChar('0')));
	
	tryToSave();
}

void MainWindow::updateCategoryList(const QStringList & categories) {
	// The filter combobox
	QString category = _filter->getCategory();
	int index = _category->findText(category);
	_category->clear();
	_category->addItem(TimekeeperFilter::AllCategories);
	_category->addItems(categories);
	if (index != -1) {
		_category->setCurrentIndex(index);
	} else {
		_category->setCurrentIndex(0);
		_filter->setCategory(TimekeeperFilter::AllCategories);
	}
	
	// The toolbar combobox
	category = _newCategory->currentText();
	_newCategory->clear();
	_newCategory->addItems(categories);
	index = _newCategory->findText(category);
	_newCategory->setCurrentIndex(index >= 0 ? index : 0);
}

QMenu* MainWindow::createPopupMenu() {
	QMenu* result = QMainWindow::createPopupMenu();
	
	result->setTitle("&Show");
	result->addSeparator();
	result->addAction(_actionShowSystray);
	
	return result;
}
