// Copyright 2007 Michiel Helvensteijn
// 
// This file is part of Timekeeper.
// 
// Timekeeper is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// Timekeeper is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Timekeeper; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

#ifndef MAINWINDOW_H
#define MAINWINDOW_H

#include "ui_mainwindow.h"

#include <QWidget>
#include <QComboBox>
#include <QToolBar>
#include <QTreeView>
#include <QSystemTrayIcon>

#include "timekeepermodel.h"
#include "timekeeperfilter.h"

class MainWindow : public QMainWindow, public Ui::MainWindow {
	Q_OBJECT
	
	public:
		
		/**
		 * Creates a standard Timekeeper main window.
		 * Ask the user if he/she wants to open a file
		 * or create a new one. Initialize stuff.
		 */
		MainWindow(QWidget * parent = 0, Qt::WFlags flags = 0);
		
		/**
		 * Destruct the main window, free up memory and do all
		 * that other good Qt stuff.
		 */
		~MainWindow();
	
	protected:
		
		/**
		 * Catch the close-event. Check if the systray icon
		 * is there or not and act accordingly. That is: close
		 * if there is no systray icon. Minimize otherwise.
		 */
		void closeEvent(QCloseEvent* event);
	
	private slots:
		
		/**
		 * Try to create a new file and clear the current model.
		 * Ask the user for the new filename.
		 * 
		 * @return True on success, false otherwise.
		 */
		bool newFile();
		
		/**
		 * Try to open a log-file.
		 * 
		 * @param filename The filename to open. If it is the empty string,
		 *                 the user is asked for the filename.
		 * @return True on success, false otherwise.
		 */
		bool openFile(QString filename = "");
		
		/**
		 * Try to import a log-file. Ask the user for the filename.
		 * 
		 * @return True on success, false otherwise.
		 */
		bool importFile();
		
		/**
		 * Take a template file and using the model, generate
		 * a report file. Ask the user for these filenames.
		 */
		void report();
		
		/**
		 * Refresh the statusbar with new information.
		 */
		void refreshStatusbar();
		
		/**
		 * Update the category-comboboxes in the filter and the activity toolbar.
		 * 
		 * @param categories The new list of categories.
		 */
		void updateCategoryList(const QStringList & categories);
		
		/**
		 * Start a new activity with the category in the
		 * activity-toolbar combobox.
		 * 
		 * If an activity is already running, an exception
		 * will be thrown and will fall through the program.
		 */
		void startActivity();
		
		/**
		 * Start a new activity and ask the user with a
		 * popup window which category.
		 * 
		 * If an activity is already running, an exception
		 * will be thrown and will fall through the program.
		 */
		void startActivitySpecify();
		
		/**
		 * Stop the currently running activity.
		 * 
		 * If no activity is running, an exception
		 * will be thrown and will fall through the program.
		 */
		void stopActivity();
		
		/**
		 * Delete the currently selected activities.
		 */
		void deleteActivities();
		
		/**
		 * Update the enabled-state of the delete button.
		 */
		void selectionChanged();
		
		/**
		 * Set the state of various indicators and buttons
		 * to indicate a new idle/running state of the program.
		 * 
		 * @param idle Whether or not the program is now idle (not running).
		 */
		void setIdleState(bool idle);
		
		/**
		 * Try to save the current model to the earlier specified file.
		 */
		void tryToSave();
		
		/**
		 * Catch a click on the systray icon. Minimize/restore this
		 * window on a single click.
		 * 
		 * @param reason The sort of click on the systray icon.
		 */
		void sysTrayActivation(QSystemTrayIcon::ActivationReason reason);
		
		/**
		 * Minimize the window if it is visible. Restore it otherwise.
		 */
		void toggleSysTrayMinimize();
		
	private: // functions
		
		/**
		 * Create and initialize various elements of the window view.
		 */
		void createView();
		
		/**
		 * Make various signal/slot connections for the operation of the window.
		 */
		void makeConnections();
		
		/**
		 * Make final initializations to the window and application before
		 * accepting any user input.
		 */
		void initialize();
		
		/**
		 * Actually start an activity with the specified category.
		 * This function does not ask any questions.
		 * 
		 * @param catgory The category of the new activity.
		 */
		void startThisActivity(const QString& category);
		
		/**
		 * Get the 'show' popup menu.
		 * 
		 * @return The 'show' popup menu.
		 */
		QMenu* createPopupMenu();
	
	private: // variables
		TimekeeperModel*  _model;
		TimekeeperFilter* _filter;
		
		QLabel*          _selection;
		QLabel*          _totalDuration;
		QComboBox*       _newCategory;
		QSystemTrayIcon* _systray;
		QMenu*           _systrayMenu;
		QMenu*           _thisPeriodMenu;
};

#endif // MAINWINDOW_H
