// Copyright 2007 Michiel Helvensteijn
// 
// This file is part of Timekeeper.
// 
// Timekeeper is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// Timekeeper is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Timekeeper; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

#include "parsernode.h"

#include <QObject>

// foreach $period $setname [in $sourceset] [of $category]
QRegExp ParserNode::_foreach("foreach\\s+(year|month|week|day|activity)\\s+([a-zA-Z_]\\w*)(?:\\s+(in)\\s+([a-zA-Z_]\\w*|\\d{2}\\/\\d{2}\\/\\d{4}\\-\\d{2}\\/\\d{2}\\/\\d{4}))?(?:\\s+(of)\\s+([a-zA-Z_]\\w*))?");

// get $setname.$dataitem[.$subdataitem]
QRegExp ParserNode::_get("get\\s+([a-zA-Z_]\\w*)\\.(?:(category)|(start)(?:\\.(d{1,4}|M{1,4}|yy|yyyy|h|hh|m|mm|s|ss))?|(end)(?:\\.(d{1,4}|M{1,4}|yy|yyyy|h|hh|m|mm|s|ss))?|(duration)(?:\\.(h|m|mm|s|ss))?|(note))");

ParserNode::ParserNode(Type type) : _type(type) {
} // intentionally empty

ParserNode::~ParserNode() {
	foreach (ParserNode* node, _children)
		delete node;
}

void ParserNode::parse(const QStringList& tokens) throw (TemplateError) {
	switch (_type) {
		case Text: case Get:
			_value = tokens[0];
			return;
		
		case Foreach:
			_value = tokens[0];
			_children.append(new ParserNode(Aggregation));
			_children.last()->parse(tokens.mid(1, tokens.count() - 2));
			return;
		
		case Aggregation:
			uint depth = 0;
			uint begin = 0;
			Type nextType = Foreach;
			bool inToken = false;
			for (int i = 0; i < tokens.count(); ++i) {
				if (depth == 0) {
					if (inToken) {
						if (!validate(tokens[i]))
							throw TemplateError(QObject::tr("There was a syntax error in the token '%1'").arg(tokens[i]));
						
						nextType = operatorType(tokens[i]);
						if (isScopeOperator(operatorType(tokens[i]))) { // Begin scope tracking
							depth = 1;
							begin = i;
						} else {
							_children.append(new ParserNode(nextType));
							_children.last()->parse(QStringList(tokens[i]));
						}
					} else {
						_children.append(new ParserNode(Text));
						_children.last()->parse(QStringList(tokens[i]));
					}
				} else if (inToken) { // Scope tracking
					if (isEnd(tokens[i])) {
						--depth;
						if (depth == 0) {
							_children.append(new ParserNode(nextType));
							_children.last()->parse(tokens.mid(begin, i - begin + 1));
						}
					} else if (isScopeOperator(operatorType(tokens[i]))) {
						++depth;
					}
				}
				inToken = !inToken;
			}
			
			if (depth > 0)
				throw TemplateError(QObject::tr("You're missing %1 ##end## token(s)."));
			
			return;
	}
}

		#include <iostream>
		using std::cerr;
		using std::endl;

void ParserNode::build(QTextStream& stream, QMap<QString, TimekeeperFilter*> sets) throw (TemplateError) {
	GetSpecs getSpecs;
	ForeachSpecs foreachSpecs;
	TimekeeperFilter *filter, *newFilter;
	
	uint duration;
	uint hours;
	uint minutes;
	uint seconds;
	
	int c, maxCount;
	bool dateRangeUsed;
	
	switch (_type) {
		case Text:
			stream << _value;
			return;
		
		case Get:
			getSpecs = getGetSpecs(_value);
			
			if (!sets.contains(getSpecs.setName))
				throw TemplateError(QObject::tr("The requested dataset '%1' does not exist in this scope.").arg(getSpecs.setName));
			
			filter = sets[getSpecs.setName];
			
			if (getSpecs.dataItem != Duration && filter->rowCount() != 1)
				throw TemplateError(QObject::tr("You asked for a data-item other than 'duration'\nand the set '%1' is not a single activity.").arg(getSpecs.setName));
			
			try {
				switch (getSpecs.dataItem) {
					case Category:
						stream << filter->onlyCategory();
						break;
					
					case Start:
						if (getSpecs.subDataItem.isEmpty())
							stream << filter->onlyStart().toString(Qt::LocaleDate);
						else
							stream << filter->onlyStart().toString(getSpecs.subDataItem);
						break;
					
					case End:
						if (getSpecs.subDataItem.isEmpty())
							stream << filter->onlyEnd().toString(Qt::LocaleDate);
						else
							stream << filter->onlyEnd().toString(getSpecs.subDataItem);
						break;
					
					case Duration:
						duration = filter->totalDuration();
						hours    = duration / (60 * 60);
						minutes  = duration / 60 % 60;
						seconds  = duration % 60;
						
						if (getSpecs.subDataItem.isEmpty())
							stream << QString("%1:%2:%3").arg(hours).arg(minutes, 2, 10, QChar('0')).arg(seconds, 2, 10, QChar('0'));
						else if (getSpecs.subDataItem == "h")
							stream << QString("%1").arg(hours);
						else if (getSpecs.subDataItem == "m")
							stream << QString("%1").arg(minutes);
						else if (getSpecs.subDataItem == "mm")
							stream << QString("%1").arg(minutes, 2, 10, QChar('0'));
						else if (getSpecs.subDataItem == "s")
							stream << QString("%1").arg(seconds);
						else // getSpecs.subDataItem == "ss"
							stream << QString("%1").arg(seconds, 2, 10, QChar('0'));
						break;
					
					case Note:
						stream << filter->onlyNote();
						break;
				}
			} catch (NotTheOnlyOne) {
				throw TemplateError();
			}
			
			return;
		
		case Foreach:
			foreachSpecs = getForeachSpecs(_value);
			
			if (foreachSpecs.source.indexOf('/') > -1) {
				dateRangeUsed = true;
				QDate from = QDate::fromString(foreachSpecs.source.section('-', 0, 0), "d/M/yyyy");
				QDate to   = QDate::fromString(foreachSpecs.source.section('-', 1, 1), "d/M/yyyy");
				
				if (!from.isValid() || !to.isValid() || from > to)
					throw TemplateError(QObject::tr("The date-range '%1' is invalid.").arg(foreachSpecs.source));
				
				filter = new TimekeeperFilter;
				filter->setSourceModel(sets["all"]);
				filter->setFrom(from);
				filter->setTo(to);
			} else {
				dateRangeUsed = false;
				if (!sets.contains(foreachSpecs.source))
					throw TemplateError(QObject::tr("The requested dataset '%1' does not exist in this scope.").arg(foreachSpecs.source));
				
				filter = sets[foreachSpecs.source];
			}
			
			if (foreachSpecs.categoryFilter.isEmpty()) {
				maxCount = filter->rowCount();
			} else {
				maxCount = filter->categoryCount(foreachSpecs.categoryFilter);
			}
			
			if (maxCount > 0) {
				newFilter = new TimekeeperFilter;
				newFilter->setSourceModel(filter);
				
				if (!foreachSpecs.categoryFilter.isEmpty())
					newFilter->setCategory(foreachSpecs.categoryFilter);
				
				if (sets.contains(foreachSpecs.setName))
					throw TemplateError(QObject::tr("The setname '%1' already exists in this scope.").arg(foreachSpecs.setName));
				
				sets.insert(foreachSpecs.setName, newFilter);
				
				switch (foreachSpecs.period) {
					case Year:
						newFilter->firstYear();
						break;
					
					case Month:
						newFilter->firstMonth();
						break;
					
					case Week:
						newFilter->firstWeek();
						break;
					
					case Day:
						newFilter->firstDay();
						break;
					
					case Activity:
						newFilter->firstActivity();
						break;
				}
				
				c = newFilter->rowCount();
				_children[0]->build(stream, sets);
				while (c < maxCount) {
					newFilter->setToNextPeriod();
					c += newFilter->rowCount();
					_children[0]->build(stream, sets);
				}
				
				delete newFilter;
			}
			if (dateRangeUsed)
				delete filter;
			
			return;
		
		case Aggregation:
			for (int i = 0; i < _children.count(); ++i) {
				_children[i]->build(stream, sets);
			}
			return;
	}
}

bool ParserNode::isScopeOperator(Type type) {
	return (type == Foreach);
}

ParserNode::Type ParserNode::operatorType(const QString& token) throw (TemplateError) {
	QString keyword = token.section(QRegExp("\\s+"), 0, 0);
	if (keyword == "foreach") {
		return Foreach;
	} else if (keyword == "get") {
		return Get;
	}
	
	throw TemplateError(QObject::tr("The keyword '%1' is not a valid token keyword.").arg(keyword));
}

bool ParserNode::isEnd(const QString& token) {
	return (token.section(QRegExp("\\s+"), 0, 0) == "end");
}

bool ParserNode::validate(const QString& token) {
	return (_foreach.exactMatch(token) || _get.exactMatch(token));
}

ParserNode::ForeachSpecs ParserNode::getForeachSpecs(const QString& token) throw (TemplateError) {
	ForeachSpecs result;
	
	if (!_foreach.exactMatch(token))
		throw TemplateError(QObject::tr("The syntax of foreach-token '%1' is invalid.").arg(token));
	
	QStringList caps = _foreach.capturedTexts();
	
	if (caps[1] == "year")  result.period = Year;  else
	if (caps[1] == "month") result.period = Month; else
	if (caps[1] == "week")  result.period = Week;  else
	if (caps[1] == "day")   result.period = Day;   else
	                        result.period = Activity;
	
	result.setName = caps[2];
	
	if (result.setName == "all")
		throw TemplateError(QObject::tr("The setname 'all' is reserved for the original dataset."));
	
	if (caps[3] == "in") {
		result.source = caps[4];
	} else {
		result.source = "all";
	}
	if (caps[5] == "of") {
		result.categoryFilter = caps[6];
	}
	
	return result;
}

ParserNode::GetSpecs ParserNode::getGetSpecs(const QString& token) throw (TemplateError) {
	GetSpecs result;
	
	if (!_get.exactMatch(token))
		throw TemplateError(QObject::tr("The syntax of get-token '%1' is invalid/").arg(token));
	
	QStringList caps = _get.capturedTexts();
	
	result.setName = caps[1];
	
	if (caps[2] == "category") {
		result.dataItem = Category;
	} else if (caps[3] == "start") {
		result.dataItem = Start;
	result.subDataItem = caps[4];
	} else if (caps[5] == "end") {
		result.dataItem = End;
	result.subDataItem = caps[6];
	} else if (caps[7] == "duration") {
		result.dataItem = Duration;
	result.subDataItem = caps[8];
	} else {
		result.dataItem = Note;
	}
	
	return result;
}

QStringList ParserNode::tokenize(const QString& string) throw (TemplateError) {
	QStringList result = string.split("##");
	
	if (result.count() % 2 == 0)
		throw TemplateError(QObject::tr("There is an uneven amount of '##'."));
	
	for (int i = 0; i < result.count(); ++i) {
		result[i].replace(QRegExp("#\\\\(\\\\*)#"), "#\\1#");
	}
	
	return result;
}
