// Copyright 2007 Michiel Helvensteijn
// 
// This file is part of Timekeeper.
// 
// Timekeeper is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// Timekeeper is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Timekeeper; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

#ifndef PARSERNODE_H
#define PARSERNODE_H

#include <QLinkedList>
#include <QStringList>
#include <QTextStream>
#include <QMap>

#include "exceptions.h"
#include "timekeeperfilter.h"

/**
 * A node in the template parse-tree.
 */
class ParserNode {
	public: // types
		
		/**
		 * The type of a node in the template parse-tree.
		 */
		enum Type {
			Text,       ///< A node with just plain text in it.
			Get,        ///< A node requesting data from the log.
			Foreach,    ///< A foreach node to browse through a timeperiod.
			Aggregation ///< An aggregation of other nodes.
		};
	
	public:
		
		/**
		 * Create a new node with type \a type .
		 * 
		 * @param type The type of the new node.
		 */
		ParserNode(Type type);
		
		/**
		 * Delete this node and the entire subtree from memory.
		 */
		~ParserNode();
		
		/**
		 * Parse the provided tokens from a timekeeper report template
		 * and build a syntax-subtree with this node as root.
		 * 
		 * @param tokens The tokens to parse.
		 * @throw TemplateError If there was a syntax error in the tokens.
		 */
		void parse(const QStringList& tokens) throw (TemplateError);
		
		/**
		 * Compile a set of model-filters into a textual report using this
		 * syntax tree. The output is sent through a textstream.
		 * The first call to this function should get a set with only the
		 * original model.
		 * 
		 * @param stream The stream to send the output to.
		 * @param sets The set of activity-sets, each with a name.
		 * @throw TemplateError If there was a symantic error in the syntax tree
		 *                      (so also in the original template file).
		 */
		void build(QTextStream& stream, QMap<QString, TimekeeperFilter*> sets) throw (TemplateError);
	
	public: // static functions
		
		/**
		 * Convert a string with template-code into a list of tokens
		 * capable of being parsed by the parse() function.
		 * 
		 * @param string The string to convert into a token-list.
		 * @param TemplateError If there is a syntax error (related to the ## sequence) in the string.
		 */
		static QStringList tokenize(const QString& string) throw (TemplateError);
	
	private: // types
		
		/**
		 * Represents a period-type that can be
		 * used to group activities in a foreach loop.
		 */
		enum Period {
			Year,
			Month,
			Week,
			Day,
			Activity
		};
		
		/**
		 * Represents the data-items you can retrieve with
		 * the 'get' command.
		 */
		enum DataItem {
			Category,
			Start,
			End,
			Duration,
			Note
		};
		
		/**
		 * Contains the information you can extract from a
		 * foreach token in the template code.
		 */
		struct ForeachSpecs {
			Period  period;
			QString setName;
			QString source;
			QString categoryFilter;
		};
		
		/**
		 * Contains the information you can extract from a
		 * 'get' token in the template code.
		 */
		struct GetSpecs {
			QString setName;
			DataItem dataItem;
			QString subDataItem;
		};
	
	private: // static functions
		
		/**
		 * Check if a certain token is one that
		 * introduces a new scope. (And needs
		 * an ##end## token.)
		 * 
		 * @param type The type of the token to test.
		 * @return Whether or not that token was a scope token.
		 */
		static bool isScopeOperator(Type type);
		
		/**
		 * Get the type of a token, given its text.
		 * 
		 * @param token The token to test.
		 * @return The type of the token.
		 * @throw TemplateError If it is not a valid token.
		 */
		static Type operatorType(const QString& token) throw (TemplateError);
		
		/**
		 * Check if the token is the ##end## token.
		 * 
		 * @param token The token to test.
		 * @return Whether or not it is the ##end## token.
		 */
		static bool isEnd(const QString& token);
		
		/**
		 * Validate the syntax of a token.
		 * 
		 * @param token The token to validate.
		 * @return Whether or not the token is a valid one.
		 */
		static bool validate(const QString& token);
		
		/**
		 * Parse a foreach token and return the information
		 * in a more accessable format.
		 * 
		 * @param token The token to parse.
		 * @return The information in the foreach-token.
		 * @throw TemplateError If it was not a valid foreach-token.
		 */
		static ForeachSpecs getForeachSpecs(const QString& token) throw (TemplateError);
		
		/**
		 * Parse a get token and return the information
		 * in a more accessable format.
		 * 
		 * @param token The token to parse.
		 * @return The information in the get-token.
		 * @throw TemplateError If it was not a valid get-token.
		 */
		static GetSpecs getGetSpecs(const QString& token) throw (TemplateError);
	
	private: // static vars
		
		static QRegExp _foreach;
		static QRegExp _get;
	
	private: // variables
		
		const Type _type;
		
		QString _value;
		QList<ParserNode*> _children;
};

#endif // PARSERNODE_H
