// Copyright 2007 Michiel Helvensteijn
// 
// This file is part of Timekeeper.
// 
// Timekeeper is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// Timekeeper is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Timekeeper; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

#include "timekeeperfilter.h"

const QString TimekeeperFilter::AllCategories = tr("All Categories");

TimekeeperFilter::TimekeeperFilter(QObject* parent) : QSortFilterProxyModel(parent), _category(TimekeeperFilter::AllCategories), _from(QDate(2000, 1, 1)), _to(QDate(2000, 1, 1)), _activityFilter(-2) {
} // intentionally empty

const QString & TimekeeperFilter::getCategory() const {
	return _category;
}

const QDate & TimekeeperFilter::getFrom() const {
	return _from;
}

const QDate & TimekeeperFilter::getTo() const {
	return _to;
}

unsigned int TimekeeperFilter::periodLength() const {
	return _from.daysTo(_to) + 1;
}

QString TimekeeperFilter::onlyCategory() const throw (NotTheOnlyOne) {
	if (rowCount() != 1)
		throw NotTheOnlyOne();
	
	return data(index(0, 0), Qt::EditRole).toString();
}

QDateTime TimekeeperFilter::onlyStart() const throw (NotTheOnlyOne) {
	if (rowCount() != 1)
		throw NotTheOnlyOne();
	
	return data(index(0, 1), Qt::EditRole).toDateTime();
}

QDateTime TimekeeperFilter::onlyEnd() const throw (NotTheOnlyOne) {
	if (rowCount() != 1)
		throw NotTheOnlyOne();
	
	return data(index(0, 2), Qt::EditRole).toDateTime();
}

uint TimekeeperFilter::totalDuration() const {
	unsigned total = 0;
	
	for (int i = 0; i < rowCount(); ++i) {
		total += QTime(0, 0, 0).secsTo(data(index(i, 3), Qt::DisplayRole).toTime());
	}
	
	return total;
}

QString TimekeeperFilter::onlyNote() const throw (NotTheOnlyOne) {
	if (rowCount() != 1)
		throw NotTheOnlyOne();
	
	return data(index(0, 4), Qt::EditRole).toString();
}

int TimekeeperFilter::categoryCount(QString category) const {
	int result = 0;
	
	for (int i = 0; i < rowCount(); ++i) {
		result += (data(index(i, 0), Qt::DisplayRole).toString() == category);
	}
	
	return result;
}

void TimekeeperFilter::setCategory(const QString & category) {
	if (category != _category) {
		_category = category;
		emit categoryChanged(category);
		
		filterChanged();
	}
}

void TimekeeperFilter::setFrom(const QDate & from) {
	if (from != _from || _activityFilter != -1) {
		_activityFilter = -1;
		if (from > _to) {
			_to = from;
			emit toChanged(_to);
		}
		
		_from = from;
		emit fromChanged(_from);
		
		filterChanged();
	}
}

void TimekeeperFilter::setTo(const QDate & to) {
	if (to != _to || _activityFilter != -1) {
		_activityFilter = -1;
		if (to < _from) {
			_from = to;
			emit fromChanged(_from);
		}
		
		_to = to;
		emit toChanged(_to);
		
		filterChanged();
	}
}

void TimekeeperFilter::setToNextPeriod() {
	if (_activityFilter == -1) {
		if (_from.dayOfYear() == 1 && _to.dayOfYear() == _from.daysInYear()) { // Shift year
			_from.setDate(_from.year() + 1, 1, 1);
			_to.setDate(_from.year(), 12, 31);
		} else if (_from.day() == 1 && _to.day() == _from.daysInMonth()) { // Shift month
			_from.setDate(_from.year() + (_from.month() == 12), _from.month() % 12 + 1, 1);
			_to.setDate(_from.year(), _from.month(), _from.daysInMonth());
		} else { // Shift day or week or other period
			int length = periodLength();
			_to = _to.addDays(length);
			_from = _from.addDays(length);
		}
		
		emit toChanged(_to);
		emit fromChanged(_from);
		filterChanged();
	} else if (_activityFilter > -1 && _activityFilter < sourceModel()->rowCount() - 1) {
		++_activityFilter;
		filterChanged();
	}
}

void TimekeeperFilter::setToPreviousPeriod() {
	if (_activityFilter == -1) {
		if (_from.dayOfYear() == 1 && _to.dayOfYear() == _from.daysInYear()) { // Shift year
			_from.setDate(_from.year() - 1, 1, 1);
			_to.setDate(_from.year(), 12, 31);
		} else if (_from.day() == 1 && _to.day() == _from.daysInMonth()) { // Shift month
			_from.setDate(_from.year() - (_from.month() == 1), (_from.month() + 10) % 12 + 1, 1);
			_to.setDate(_from.year(), _from.month(), _from.daysInMonth());
		} else { // Shift day or week or other period
			int length = periodLength();
			_from = _from.addDays(-length);
			_to = _to.addDays(-length);
		}
		
		emit fromChanged(_from);
		emit toChanged(_to);
		filterChanged();
	} else if (_activityFilter > 0) {
		--_activityFilter;
		filterChanged();
	}
}

void TimekeeperFilter::setAll() {
	_activityFilter = -2;
}


void TimekeeperFilter::thisDay() {
	_activityFilter = -1;
	QDate c = QDate::currentDate();
	
	_from = c;
	emit fromChanged(_from);
	_to = c;
	emit toChanged(_to);
	
	filterChanged();
}

void TimekeeperFilter::thisWeek() {
	_activityFilter = -1;
	QDate c = QDate::currentDate();
	
	_from = c.addDays(1 - c.dayOfWeek());
	emit fromChanged(_from);
	_to = c.addDays(7 - c.dayOfWeek());
	emit toChanged(_to);
	
	filterChanged();
}

void TimekeeperFilter::thisMonth() {
	_activityFilter = -1;
	QDate c = QDate::currentDate();
	
	_from = c.addDays(1 - c.day());
	emit fromChanged(_from);
	_to = c.addDays(c.daysInMonth() - c.day());
	emit toChanged(_to);
	
	filterChanged();
}

void TimekeeperFilter::thisYear() {
	_activityFilter = -1;
	QDate c = QDate::currentDate();
	
	_from = c.addDays(1 - c.dayOfYear());
	emit fromChanged(_from);
	_to = c.addDays(c.daysInYear() - c.dayOfYear());
	emit toChanged(_to);
	
	filterChanged();
}

void TimekeeperFilter::firstDay() {
	_activityFilter = -1;
	QDate c = getFirstTimestamp().date();
	
	_from = c;
	emit fromChanged(_from);
	_to = c;
	emit toChanged(_to);
	
	filterChanged();
}

void TimekeeperFilter::firstWeek() {
	_activityFilter = -1;
	QDate c = getFirstTimestamp().date();
	
	_from = c.addDays(1 - c.dayOfWeek());
	emit fromChanged(_from);
	_to = c.addDays(7 - c.dayOfWeek());
	emit toChanged(_to);
	
	filterChanged();
}

void TimekeeperFilter::firstMonth() {
	_activityFilter = -1;
	QDate c = getFirstTimestamp().date();
	
	_from = c.addDays(1 - c.day());
	emit fromChanged(_from);
	_to = c.addDays(c.daysInMonth() - c.day());
	emit toChanged(_to);
	
	filterChanged();
}

void TimekeeperFilter::firstYear() {
	_activityFilter = -1;
	QDate c = getFirstTimestamp().date();
	
	_from = c.addDays(1 - c.dayOfYear());
	emit fromChanged(_from);
	_to = c.addDays(c.daysInYear() - c.dayOfYear());
	emit toChanged(_to);
	
	filterChanged();
}

void TimekeeperFilter::firstActivity() {
	_activityFilter = 0;
	
	filterChanged();
}

bool TimekeeperFilter::filterAcceptsRow(int sourceRow, const QModelIndex& sourceParent) const {
	if (_activityFilter == -2) {
		return true;
	} else if (_activityFilter > -1) {
		return (sourceRow == _activityFilter);
	} else {
		QModelIndex categoryIndex = sourceModel()->index(sourceRow, 0, sourceParent);
		QString     category      = sourceModel()->data(categoryIndex).toString();
		
		QModelIndex startIndex = sourceModel()->index(sourceRow, 1, sourceParent);
		QDateTime   start      = sourceModel()->data(startIndex, Qt::EditRole).toDateTime();
		
		QModelIndex endIndex = sourceModel()->index(sourceRow, 2, sourceParent);
		QDateTime   end      = sourceModel()->data(endIndex, Qt::EditRole).toDateTime();
		
		bool categoryMatches = (_category == AllCategories ||
								_category == category);
		bool dateMatches     = ((_from <= start.date() && start.date() <= _to) ||
								(_from <= end.date()   && end.date()   <= _to));
		
		return categoryMatches && dateMatches;
	}
}

QDateTime TimekeeperFilter::getFirstTimestamp() const {
	QDateTime result;
	
	if (sourceModel()->rowCount() > 0) {
		result = sourceModel()->data(sourceModel()->index(0, 1), Qt::EditRole).toDateTime();
	}
	for (int i = 1; i < sourceModel()->rowCount(); ++i) {
		if (result > sourceModel()->data(sourceModel()->index(i, 1), Qt::EditRole).toDateTime()) {
			result = sourceModel()->data(sourceModel()->index(i, 1), Qt::EditRole).toDateTime();
		}
	}
	
	return result;
}
