// Copyright 2007 Michiel Helvensteijn
// 
// This file is part of Timekeeper.
// 
// Timekeeper is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// Timekeeper is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Timekeeper; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

#ifndef TIMEKEEPERFILTER_H
#define TIMEKEEPERFILTER_H

#include <QSortFilterProxyModel>
#include <QString>
#include <QDate>

#include "exceptions.h"

/**
 * A class to filter output from the TimekeeperModel class,
 * in the Qt Model-View architecture. Provides an easy to
 * use interface to filter by timeperiod (or single activity)
 * or by category.
 * 
 * @see TimekeeperModel
 */
class TimekeeperFilter : public QSortFilterProxyModel {
	Q_OBJECT
	
	public:
		
		/// The string representing the 'all categories' filter.
		static const QString AllCategories;
		
	public:
		
		/**
		 * Creates a filter that initially lets all activities
		 * through. No category filter, no date filter.
		 * 
		 * @param parent The QObject parent of this filter.
		 */
		TimekeeperFilter(QObject* parent = NULL);
		
		/**
		 * Get the category that is being filtered by.
		 * 
		 * @return The category that is being filtered by.
		 */
		const QString& getCategory() const;
		
		/**
		 * Get the 'from' date. If an activity has a date (from or to)
		 * equal or higher than this date, it may be included.
		 * 
		 * @return The 'from' date.
		 */
		const QDate& getFrom() const;
		
		/**
		 * Get the 'to' date. If an activity has a date (from or to)
		 * equal or lower than this date, it may be included.
		 * 
		 * @return The 'to' date.
		 */
		const QDate& getTo() const;
		
		/**
		 * Get the length of the filter-period in seconds.
		 * 
		 * @return The length of the filter-period in seconds.
		 */
		uint periodLength() const;
		
		/**
		 * Get the only category that is let through the filter,
		 * if there is exactly one such category.
		 * 
		 * @return The only category.
		 * @throw NotTheOnlyOne If there is not exactly one category.
		 */
		QString onlyCategory() const throw (NotTheOnlyOne);
		
		/**
		 * Get the only start-date/time that is let through the filter,
		 * if there is exactly one such start-date/time.
		 * 
		 * @return The only start-date/time.
		 * @throw NotTheOnlyOne If there is not exactly one start-date/time.
		 */
		QDateTime onlyStart() const throw (NotTheOnlyOne);
		
		/**
		 * Get the only end-date/time that is let through the filter,
		 * if there is exactly one such end-date/time.
		 * 
		 * @return The only end-date/time.
		 * @throw NotTheOnlyOne If there is not exactly one end-date/time.
		 */
		QDateTime onlyEnd() const throw (NotTheOnlyOne);
		
		/**
		 * Get the sum of the durations of all activities that are
		 * let through the filter.
		 * 
		 * @return The total duration of everything let through.
		 */
		uint totalDuration() const;
		
		/**
		 * Get the only note that is let through the filter,
		 * if there is exactly one such note.
		 * 
		 * @return The only note.
		 * @throw NotTheOnlyOne If there is not exactly one note.
		 */
		QString onlyNote() const throw (NotTheOnlyOne);
		
		/**
		 * Get the number of rows with the category \a category .
		 * 
		 * @param category The category to ge the number of rows of.
		 * @return The number of rows with the specified category.
		 */
		int categoryCount(QString category) const;
		
	public slots:
		
		/**
		 * Set the category to filter by.
		 * 
		 * @param category The category to filter by.
		 */
		void setCategory(const QString & category);
		
		/**
		 * Set the 'from' date to filter by.
		 * 
		 * @param from The 'from' date to filter by.
		 */
		void setFrom(const QDate & from);
		
		/**
		 * Set the 'to' date to filter by.
		 * 
		 * @param to The 'to' date to filter by.
		 */
		void setTo(const QDate & to);
		
		/**
		 * Set the filter to the next period. This function
		 * recognizes boundries of days, weeks, months and years.
		 */
		void setToNextPeriod();
		
		/**
		 * Set the filter to the previous period. This function
		 * recognizes boundries of activities, days, weeks, months 
		 * and years.
		 */
		void setToPreviousPeriod();
		
		/**
		 * Set the filter to accept all activities.
		 * This removes the category and date filters.
		 */
		void setAll();
		
		/**
		 * Set the date-filter to the current day.
		 */
		void thisDay();
		
		/**
		 * Set the date-filter to the current week.
		 */
		void thisWeek();
		
		/**
		 * Set the date-filter to the current month.
		 */
		void thisMonth();
		
		/**
		 * Set the date-filter to the current year.
		 */
		void thisYear();
		
		/**
		 * Set the date-filter to the first day where the
		 * source model has any activity.
		 */
		void firstDay();
		
		/**
		 * Set the date-filter to the first week where the
		 * source model has any activity.
		 */
		void firstWeek();
		
		/**
		 * Set the date-filter to the first month where the
		 * source model has any activity.
		 */
		void firstMonth();
		
		/**
		 * Set the date-filter to the first year where the
		 * source model has any activity.
		 */
		void firstYear();
		
		/**
		 * Set the date-filter to the first activity of
		 * the source model.
		 */
		void firstActivity();
		
	signals:
		
		/**
		 * Emitted when the filter-category has changed.
		 * 
		 * @param category The new filter-category.
		 */
		void categoryChanged(const QString & category);
		
		/**
		 * Emitted when the 'from' date filter has changed.
		 * 
		 * @param from The new 'from' date.
		 */
		void fromChanged(const QDate & from);
		
		/**
		 * Emitted when the 'to' date filter has changed.
		 * 
		 * @param to The new 'to' date.
		 */
		void toChanged(const QDate & to);
		
	protected:
		
		/**
		 * Checks if a certain row from the source model is accepted or not.
		 * Reimplemented from QSortFilterProxyModel.
		 */
		bool filterAcceptsRow(int sourceRow, const QModelIndex& sourceParent) const;
	
	private: // functions
		
		/**
		 * Get the first date/time from any activity in the
		 * source-model.
		 */
		QDateTime getFirstTimestamp() const;
	
	private: // variables
		
		QString _category;
		QDate _from;
		QDate _to;
		
		int _activityFilter;
};

#endif // TIMEKEEPERFILTER_H
