// Copyright 2007 Michiel Helvensteijn
// 
// This file is part of Timekeeper.
// 
// Timekeeper is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// Timekeeper is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Timekeeper; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

#include "timekeepermodel.h"

#include <QFile>
#include <QBrush>
#include <QColor>
#include <QtXml>

TimekeeperModel::TimekeeperModel(QObject* parent) : QAbstractTableModel(parent) {
	// Setup timer
	_timer.setInterval(1000); // Set update interval to 1 second
	_timer.setSingleShot(false); // Set to repeditive
	connect(&_timer, SIGNAL(timeout()), this, SLOT(updateRunningActivity()));
}

int TimekeeperModel::rowCount(const QModelIndex& /* parent */) const {
	return _activities.count();
}

int TimekeeperModel::columnCount(const QModelIndex& /* parent */) const {
	return 5;
}

QVariant TimekeeperModel::data(const QModelIndex& index, int role) const {
	if (!index.isValid())
		return QVariant();
	
	bool running = (index.row() == _activities.count() - 1 && activityRunning());
	
	if (role == Qt::DisplayRole) {
		switch (index.column()) {
			case 0: return _activities[index.row()].getCategory();
			case 1: return _activities[index.row()].getStart().toString(Qt::LocaleDate);
			case 2: return _activities[index.row()].getEnd().toString(Qt::LocaleDate);
			case 3: return _activities[index.row()].getDuration().toString("h:mm:ss");
			case 4: return _activities[index.row()].getNote();
		}
	} else if (role == Qt::EditRole) {
		switch (index.column()) {
			case 0: return _activities[index.row()].getCategory();
			case 1: return _activities[index.row()].getStart();
			case 2: return _activities[index.row()].getEnd();
			case 3: return _activities[index.row()].getDuration();
			case 4: return _activities[index.row()].getNote();
		}
	} else if (role == Qt::DecorationRole && running && index.column() == 3) {
		return QIcon("images/running_32x32.png");
	} else if (role == Qt::BackgroundRole && running) {
		return QBrush(QColor(180, 255, 180));
	} else if (role == Qt::TextAlignmentRole && index.column() >= 1 && index.column() <= 3) {
		return Qt::AlignRight;
	}
	
	return QVariant();
}

bool TimekeeperModel::setData(const QModelIndex& index, const QVariant& value, int role) {
	if (!index.isValid())
		return false;
	
	bool running = (index.row() == _activities.count() - 1 && activityRunning());
	
	if (role == Qt::EditRole) {
		switch (index.column()) {
			case 0:
				_activities[index.row()].setCategory(value.toString());
				addCategory(value.toString());
				break;
			case 1:
				if (running) {
					return false;
				} else {
					_activities[index.row()].setStart(value.toDateTime());
					emit dataChanged(createIndex(index.row(), 3), createIndex(index.row(), 3));
				}
				break;
			case 2:
				if (running) {
					return false;
				} else {
					_activities[index.row()].setEnd(value.toDateTime());
					emit dataChanged(createIndex(index.row(), 3), createIndex(index.row(), 3));
				}
				break;
			case 3:
				if (running) {
					return false;
				} else {
					_activities[index.row()].setDuration(value.toTime());
					emit dataChanged(createIndex(index.row(), 2), createIndex(index.row(), 2));
				}
				break;
			case 4:
				_activities[index.row()].setNote(value.toString());
				break;
		}
		emit dataChanged(index, index);
		
		return QAbstractItemModel::setData(index, value, role);
	}
	
	return false;
}

QVariant TimekeeperModel::headerData(int section, Qt::Orientation orientation, int role) const {
	if (orientation == Qt::Horizontal && role == Qt::DisplayRole) {
		switch (section) {
			case 0: return "Category";
			case 1: return "Start";
			case 2: return "End";
			case 3: return "Duration";
			case 4: return "Note";
		}
	}
	
	return QVariant();
}

Qt::ItemFlags TimekeeperModel::flags(const QModelIndex& index) const {
	Qt::ItemFlags flags = QAbstractItemModel::flags(index);
	
	if (index.row() < _activities.count() - activityRunning() || index.column() == 0 || index.column() == 4) {
		flags |= Qt::ItemIsEditable;
	}
	
	return flags;
}

void TimekeeperModel::addActivity(const QString & category, const QDateTime & start, const QDateTime & end, const QString & note) {
	beginInsertRows(QModelIndex(), _activities.count() - activityRunning(), _activities.count() - activityRunning());
	_activities.insert(_activities.count() - activityRunning(), Activity(category, start, end, note));
	
	endInsertRows();
}

void TimekeeperModel::startActivity(const QString & category) throw (ActivityIsRunning) {
	if (activityRunning())
		throw ActivityIsRunning();
	
	beginInsertRows(QModelIndex(), _activities.count(), _activities.count());
	_activities.append(Activity(category, QDateTime::currentDateTime(), QDateTime::currentDateTime()));
	_timer.start();
	endInsertRows();
}

void TimekeeperModel::stopCurrentActivity(const QDateTime & end) throw (NoActivityRunning) {
	if (!activityRunning())
		throw NoActivityRunning();
	
	_activities.last().setEnd(end);
	_timer.stop();
	emit dataChanged(createIndex(_activities.count() - 1, 2), createIndex(_activities.count() - 1, 3));
}

bool TimekeeperModel::removeRows(int row, int count, const QModelIndex& /* parent */) {
	if (row < 0 || row + count > _activities.count())
		return false;
	
	beginRemoveRows(QModelIndex(), row, row + count - 1);
	if (row + count == _activities.count() && activityRunning()) {
		_timer.stop();
	}
	_activities.remove(row, count);
	endRemoveRows();
	
	return true;
}

bool TimekeeperModel::activityRunning() const {
	return _timer.isActive();
}

QString TimekeeperModel::runningCategory() const throw (NoActivityRunning) {
	if (!activityRunning())
		throw NoActivityRunning();
	
	return _activities.last().getCategory();
}

void TimekeeperModel::clear() {
	internalClear();
	reset();
}

void TimekeeperModel::load() throw (NoSuchFile, NoPermission, InvalidFile) {
	importAssert(getFilename());
	internalClear();
	internalImport(getFilename());
	reset();
}

void TimekeeperModel::import(QString filename) throw (NoSuchFile, NoPermission, InvalidFile) {
	importAssert(filename);
	internalImport(filename);
	reset();
}

void TimekeeperModel::save() const throw (NoPermission, UnknownError) {
	QDomDocument doc("log");
	
	QDomElement root = doc.createElement("timekeeperlog");
	doc.appendChild(root);
	
	////////////////////////
	////// ACTIVITIES //////
	////////////////////////
	
	QDomElement activities = doc.createElement("activities");
	root.appendChild(activities);
	
	for (int i = 0; i < _activities.count() - activityRunning(); ++i) {
		QDomElement activity = doc.createElement("activity");
		activities.appendChild(activity);
		
		// category
		
		QDomElement category = doc.createElement("category");
		activity.appendChild(category);
		
		QDomText tCategory = doc.createTextNode(_activities[i].getCategory());
		category.appendChild(tCategory);
		
		// from
		
		QDomElement from = doc.createElement("from");
		activity.appendChild(from);
		
		QDomText tFrom = doc.createTextNode(_activities[i].getStart().toString());
		from.appendChild(tFrom);
		
		// to
		
		QDomElement to = doc.createElement("to");
		activity.appendChild(to);
		
		QDomText tTo = doc.createTextNode(_activities[i].getEnd().toString());
		to.appendChild(tTo);
		
		// note
		
		if (!_activities[i].getNote().isEmpty()) {
			QDomElement note = doc.createElement("note");
			activity.appendChild(note);
			
			QDomText tNote = doc.createTextNode(_activities[i].getNote());
			note.appendChild(tNote);
		}
	}
	
	if (activityRunning()) {
		QDomElement activity = doc.createElement("activity");
		activities.appendChild(activity);
		
		// category
		
		QDomElement category = doc.createElement("category");
		activity.appendChild(category);
		
		QDomText tCategory = doc.createTextNode(_activities.last().getCategory());
		category.appendChild(tCategory);
		
		// from
		
		QDomElement from = doc.createElement("from");
		activity.appendChild(from);
		
		QDomText tFrom = doc.createTextNode(_activities.last().getStart().toString());
		from.appendChild(tFrom);
		
		// note
		
		if (!_activities.last().getNote().isEmpty()) {
			QDomElement note = doc.createElement("note");
			activity.appendChild(note);
			
			QDomText tNote = doc.createTextNode(_activities.last().getNote());
			note.appendChild(tNote);
		}
	}
	
	QFile file(getFilename());
	
	if (!file.open(QIODevice::WriteOnly))
		throw NoPermission();
	
	QString data = doc.toString(4);
	
	if (file.write(data.toStdString().c_str(), data.length()) == -1) {
		file.close();
		throw UnknownError();
	}
	
	file.close();
}

void TimekeeperModel::addCategory(QString category) {
	if (!_categories.contains(category)) {
		_categories.append(category);
	}
	
	emit categoriesChanged(_categories);
}

void TimekeeperModel::removeCategory(QString category) {
	_categories.removeAll(category);
	
	emit categoriesChanged(_categories);
}

const QStringList & TimekeeperModel::getCategories() const {
	return _categories;
}

void TimekeeperModel::internalClear() {
	if (activityRunning())
		_timer.stop();
	
	_categories.clear();
	_activities.clear();
}

void TimekeeperModel::internalImport(QString filename) throw (NoSuchFile, NoPermission, InvalidFile) {
	QFile file(filename);
	
	if (!file.exists())
		throw NoSuchFile();
	
	if (!file.open(QIODevice::ReadOnly))
		throw NoPermission();
	
	QDomDocument doc("model");
	
	if (!doc.setContent(&file)) {
		file.close();
		throw InvalidFile();
	}
	
	file.close();
	
	  /////////////////////////////////////////////////
	 //////////// Make the actual changes ////////////
	/////////////////////////////////////////////////
	
	QDomElement root = doc.documentElement();
	
	////////////////////////
	////// ACTIVITIES //////
	////////////////////////
	
	QDomElement activities = root.firstChildElement("activities");
	
	Activity runningOne;
	
	bool runningActivityFound = false;
	
	for (QDomNode actNode = activities.firstChild(); !actNode.isNull(); actNode = actNode.nextSibling()) {
		QDomElement activity = actNode.toElement();
		
		QDomElement category = activity.firstChildElement("category");
		QString sCategory = category.text();
		addCategory(sCategory);
		
		QDomElement from = activity.firstChildElement("from");
		QDateTime dFrom = QDateTime::fromString(from.text());
		
		QDomElement note = activity.firstChildElement("note");
		QString sNote = (note.isNull() ? QString("") : note.text());
		
		QDomElement to = activity.firstChildElement("to");
		if (to.isNull()) {
			runningActivityFound = true;
			runningOne = Activity(sCategory, dFrom, QDateTime::currentDateTime(), sNote);
		} else {
			_activities.append(Activity(sCategory, dFrom, QDateTime::fromString(to.text()), sNote));
		}
	}
	
	if (runningActivityFound) {
		_activities.append(runningOne);
		_timer.start();
	}
}

void TimekeeperModel::importAssert(QString filename) throw (NoSuchFile, NoPermission, InvalidFile) {
	QFile file(filename);
	
	if (!file.exists())
		throw NoSuchFile();
	
	if (!file.open(QIODevice::ReadOnly))
		throw NoPermission();
	
	QDomDocument doc("model");
	
	if (!doc.setContent(&file)) {
		file.close();
		throw InvalidFile();
	}
	
	file.close();
	
	  /////////////////////////////////////////////////////////////////////////////////////////
	 //////////// See if an exception should be thrown, without changing anything ////////////
	/////////////////////////////////////////////////////////////////////////////////////////
	
	QDomElement root = doc.documentElement();
	
	////////////////////////
	////// ACTIVITIES //////
	////////////////////////
	
	bool runningActivityFound = false;
	
	QDomElement activities = root.firstChildElement("activities");
	if (activities.isNull())
		throw InvalidFile();
	
	for (QDomNode actNode = activities.firstChild(); !actNode.isNull(); actNode = actNode.nextSibling()) {
		QDomElement activity = actNode.toElement();
		
		QDomElement category = activity.firstChildElement("category");
		if (category.isNull())
			throw InvalidFile();
		
		QDomElement from = activity.firstChildElement("from");
		if (from.isNull())
			throw InvalidFile();
		
		// There does not have to be a note
		
		QDomElement to = activity.firstChildElement("to");
		if (to.isNull()) {
			if (runningActivityFound || activityRunning())
				throw InvalidFile();
			runningActivityFound = true;
		}
	}
}

void TimekeeperModel::updateRunningActivity() throw (NoActivityRunning) {
	if (!activityRunning())
		throw NoActivityRunning();
	
	_activities.last().setEnd(QDateTime::currentDateTime());
	emit dataChanged(createIndex(_activities.count() - 1, 2), createIndex(_activities.count() - 1, 3));
}
