// Copyright 2007 Michiel Helvensteijn
// 
// This file is part of Timekeeper.
// 
// Timekeeper is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// Timekeeper is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Timekeeper; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

#ifndef TIMEKEEPERMODEL_H
#define TIMEKEEPERMODEL_H

#include <QAbstractTableModel>
#include <QModelIndex>
#include <QString>
#include <QDateTime>
#include <QTimer>
#include <QVector>
#include <QIcon>

#include "activity.h"
#include "exceptions.h"
#include "logfilehandler.h"

/**
 * The timekeeper model stores all activities from a log-file
 * and has the interface to connect through the Qt
 * Model-View architecture to a flat QTreeView.
 * 
 * It also saves the list of previously used activities and it
 * has the interface to save, load and import log-files.
 * 
 * @see TimekeeperFilter
 */
class TimekeeperModel : public QAbstractTableModel, public LogfileHandler {
	Q_OBJECT
	
	public:
		
		/**
		 * Creates a new model without any activities in it.
		 */
		TimekeeperModel(QObject* parent = NULL);
		
		/**
		 * Get the number of activities in this model.
		 * 
		 * @return The number of activities in this model.
		 */
		int rowCount(const QModelIndex& /* parent */) const;
		
		/**
		 * Get the number of data-items in an activity.
		 * 
		 * @return 5. Namely: Category, Start, End, Duration, Note
		 */
		int columnCount(const QModelIndex& /* parent */) const;
		
		/**
		 * Get a piece of data from the model.
		 * 
		 * @param index The piece of data to retrieve.
		 * @param role The role of the data to retrieve.
		 * @return The data.
		 */
		QVariant data(const QModelIndex& index, int role) const;
		
		/**
		 * Change a piece of data in the model.
		 * 
		 * @param index The piece of data to change.
		 * @param value The new value of that data-location.
		 * @param role The role of the data.
		 * @return True on success. False on failure.
		 */
		bool setData(const QModelIndex& index, const QVariant& value, int role);
		
		/**
		 * Get data for a column in the header.
		 * 
		 * @param section The section of the header.
		 * @param orientation Should be Qt::Horizontal.
		 * @param role The role of the data. Should be Qt::DisplayRole.
		 */
		QVariant headerData(int section, Qt::Orientation orientation, int role) const;
		
		/**
		 * Get the flags for a piece of data in the model.
		 * 
		 * @param index The piece of data to get the flags for.
		 * @return The flags for that piece of data.
		 */
		Qt::ItemFlags flags(const QModelIndex& index) const;
		
		/**
		 * Remove \a count rows starting with \a row from the model.
		 * 
		 * @param row The starting-row to remove.
		 * @param count The number of rows to remove.
		 * @return True on success. False on failure.
		 */
		bool removeRows(int row, int count, const QModelIndex& /* parent */);
		
		/**
		 * Add a completed activity to the model.
		 * 
		 * @param category The category of the new activity.
		 * @param start The start-date/time of the new activity.
		 */
		void addActivity(const QString & category, const QDateTime & start, const QDateTime & end, const QString & note = QString());
		
		/**
		 * Start a new activity.
		 * From this moment on, a clock will automatically update the new
		 * activity every second, to create a running clock in the view.
		 * 
		 * @param category The category of the new activity.
		 * @throw ActivityIsRunning If there is already an activity running.
		 */
		void startActivity(const QString & category) throw (ActivityIsRunning);
		
		/**
		 * Stop the currently running activity.
		 * 
		 * @param end The end-time of the activity.
		 * @throw NoActivityRunning If no activity is currently running.
		 */
		void stopCurrentActivity(const QDateTime & end = QDateTime::currentDateTime()) throw (NoActivityRunning);
		
		/**
		 * Check if an activity is running.
		 * 
		 * @return Whether or not an activity is running.
		 */
		bool activityRunning() const;
		
		/**
		 * Get the category of the running activity.
		 * 
		 * @return The category of the running activity.
		 * @throw NoActivityRunning If no activity is currently running.
		 */
		QString runningCategory() const throw (NoActivityRunning);
		
		/**
		 * Clear the entire model and let the connected views know.
		 */
		void clear();
		
		/**
		 * Load the file of which the filename has been given to this object
		 * and let the connected views know.
		 * 
		 * @throw NoSuchFile If no such file exists.
		 * @throw NoPermission If you have no permission to read this file.
		 * @throw InvalidFile If this file does not have the correct timekeeper log format.
		 */
		void load() throw (NoSuchFile, NoPermission, InvalidFile);
		
		/**
		 * Import the specified file into the current model and let
		 * the connected views know.
		 * 
		 * @param filename The file to import.
		 * @throw NoSuchFile If no such file exists.
		 * @throw NoPermission If you have no permission to read this file.
		 * @throw InvalidFile If this file does not have the correct timekeeper log format.
		 */
		void import(QString filename) throw (NoSuchFile, NoPermission, InvalidFile);
		
		/**
		 * Save this model to the filname that has ben given to this object.
		 * 
		 * @throw NoPermission If you have no permission to write to this file.
		 * @throw UnknownError If an unknown error occurred.
		 */
		void save() const throw (NoPermission, UnknownError);
		
		/**
		 * Add a category to the list of previously used categories.
		 * 
		 * @param category The category to add.
		 */
		void addCategory(QString category);
		
		/**
		 * Remove a category from the list of previously used categories.
		 * 
		 * @param category The category to remove.
		 */
		void removeCategory(QString category);
		
		/**
		 * Get the list of previously used categories.
		 * 
		 * @return The list of previously used categories.
		 */
		const QStringList & getCategories() const;
	
	signals:
		
		/**
		 * Emitted if the list of previously used categories has changed.
		 * 
		 * @param categories The new list of categories.
		 */
		void categoriesChanged(const QStringList & categories);
	
	private:
		
		/**
		 * Clear the model without notifying connected views.
		 */
		void internalClear();
		
		/**
		 * Import a file into this model without notifying connected views.
		 * 
		 * @param filename The file to import.
		 * @throw NoSuchFile If no such file exists.
		 * @throw NoPermission If you have no permission to read this file.
		 * @throw InvalidFile If this file does not have the correct timekeeper log format.
		 */
		void internalImport(QString filename) throw (NoSuchFile, NoPermission, InvalidFile);
		
		/**
		 * Check if an exception should be thrown upon importing a file, without actually
		 * importing it.
		 * 
		 * @param filename The file to import.
		 * @throw NoSuchFile If no such file exists.
		 * @throw NoPermission If you have no permission to read this file.
		 * @throw InvalidFile If this file does not have the correct timekeeper log format.
		 */
		void importAssert(QString filename) throw (NoSuchFile, NoPermission, InvalidFile);
	
	private slots:
		
		/**
		 * Update the end-time and duration of the running activity.
		 * 
		 * @throw NoActivityRunning If no activity is running.
		 */
		void updateRunningActivity() throw (NoActivityRunning);
	
	private:
		QVector<Activity> _activities;
		QTimer _timer;
		
		QStringList _categories;
};

#endif // TIMEKEEPERMODEL_H
